package com.tencent.igame.tip_flutter_image_view;


import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.drawable.Drawable;
import android.opengl.GLES20;
import android.opengl.GLUtils;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.util.Log;

import com.bumptech.glide.Glide;
import com.bumptech.glide.request.target.CustomTarget;
import com.bumptech.glide.request.transition.Transition;

import java.io.File;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.Map;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import io.flutter.plugin.common.EventChannel;


public class ImageRenderWorker implements OpenGLRenderer.Worker {

    private static final String TAG = "ImageRenderWorker";
    private Context context;
    private String imageUrl;
    private BitmapProvider bitmapProvider;
    private int width;
    private int height;
    private boolean centerCrop;
    private int textureId;
    private HandlerThread handlerThread = new HandlerThread("Handler Thread");

    //位置
    private FloatBuffer vertexBuffer;
    //纹理
    private FloatBuffer textureBuffer;
    //顶点坐标
    static float vertexData[] = {   // in counterclockwise order:
            -1f, -1f, 0.0f, // bottom left
            1f, -1f, 0.0f, // bottom right
            -1f, 1f, 0.0f, // top left
            1f, 1f, 0.0f,  // top right
    };

    //纹理坐标  对应顶点坐标  与之映射
    static float textureData[] = {   // in counterclockwise order:
            0f, 1f, 0.0f, // bottom left
            1f, 1f, 0.0f, // bottom right
            0f, 0f, 0.0f, // top left
            1f, 0f, 0.0f,  // top right
    };

    //每一次取点的时候取几个点
    static final int COORDS_PER_VERTEX = 3;

    private final int vertexCount = vertexData.length / COORDS_PER_VERTEX;
    //每一次取的总的点 大小
    private final int vertexStride = COORDS_PER_VERTEX * 4; // 4 bytes per vertex

    private int program;
    private int avPosition;
    //纹理位置
    private int afPosition;
    private Handler mWorkerHandler;
//    private EventChannel.EventSink sink;

    public ImageRenderWorker(Context context, int textureId, String imageUrl, int width, int height, boolean centerCrop, Map<String, EventChannel.EventSink> sinkMap) {
        this.context = context;
        this.textureId = textureId;
        this.imageUrl = imageUrl;
        this.width = width;
        this.height = height;
        this.centerCrop = centerCrop;
//        this.sink = sinkMap.get(imageUrl);
    }

    @Override
    public void onCreate() {
        handlerThread.start();
        mWorkerHandler = new Handler(handlerThread.getLooper());
        initBitmapProvider();
        createProgram();
        createTexture();
    }

    /**
     * 创建纹理
     */
    private void createTexture() {
        if (program > 0) {
            //获取顶点坐标字段
            avPosition = GLES20.glGetAttribLocation(program, "av_Position");
            //获取纹理坐标字段
            afPosition = GLES20.glGetAttribLocation(program, "af_Position");
            //绑定纹理
            Log.e("ImageRenderWorker", "glBindTexture() called textureId " + textureId);
            //fixme 这里其实就是flutter中texture控件的那个textureId，这里比较关键
            GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textureId);
            //环绕（超出纹理坐标范围）  （s==x t==y GL_REPEAT 重复）
            GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_S, GLES20.GL_REPEAT);
            GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_T, GLES20.GL_REPEAT);
            //过滤（纹理像素映射到坐标点）  （缩小、放大：GL_LINEAR线性）
            GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MIN_FILTER, GLES20.GL_LINEAR);
            GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MAG_FILTER, GLES20.GL_LINEAR);
            //设置纹理为2d图片
            updateTexture();
        }
    }

    /**
     * 主要是为了渲染动图
     */
    private void updateTexture() {
        Bitmap bitmap = bitmapProvider.get();
        //设置纹理为2d图片
        if (bitmap != null && !bitmap.isRecycled()) {
            GLUtils.texImage2D(GLES20.GL_TEXTURE_2D, 0, bitmap, 0);
        }
    }

    private void createProgram() {
        vertexBuffer = ByteBuffer.allocateDirect(vertexData.length * 4)
                .order(ByteOrder.nativeOrder())
                .asFloatBuffer()
                .put(vertexData);
        vertexBuffer.position(0);

        textureBuffer = ByteBuffer.allocateDirect(textureData.length * 4)
                .order(ByteOrder.nativeOrder())
                .asFloatBuffer()
                .put(textureData);
        textureBuffer.position(0);

        String vertexSource = ShaderUtil.readRawTxt(context, R.raw.vertex_shader);
        String fragmentSource = ShaderUtil.readRawTxt(context, R.raw.fragment_shader);
        program = ShaderUtil.createProgram(vertexSource, fragmentSource);
        Log.i("ImageRenderWorker", "onCreate() called  program  " + program);
    }

    @Override
    public boolean onDraw() {
        //底色
        GLES20.glClearColor(0.5f, 0.5f, 0.5f, 1f);
        GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);
        //使用程序
        GLES20.glUseProgram(program);
        GLES20.glEnableVertexAttribArray(avPosition);
        GLES20.glEnableVertexAttribArray(afPosition);
        //设置顶点位置值
        GLES20.glVertexAttribPointer(avPosition, COORDS_PER_VERTEX, GLES20.GL_FLOAT, false, vertexStride, vertexBuffer);
        //设置纹理位置值
        GLES20.glVertexAttribPointer(afPosition, COORDS_PER_VERTEX, GLES20.GL_FLOAT, false, vertexStride, textureBuffer);
        //绘制 GLES20.GL_TRIANGLE_STRIP:复用坐标
        GLES20.glDrawArrays(GLES20.GL_TRIANGLE_STRIP, 0, vertexCount);
        GLES20.glDisableVertexAttribArray(avPosition);
        GLES20.glDisableVertexAttribArray(afPosition);
        bitmapProvider.next();
        updateTexture();
        return true;
    }


    /**
     * 初始化Provider
     */
    private void initBitmapProvider() {
        bitmapProvider = new NormalBitmapProvider(context, imageUrl, width, height, centerCrop);
        Glide.with(context)
                .downloadOnly()
                .load(imageUrl).into(new CustomTarget<File>() {
            @Override
            public void onResourceReady(@NonNull File resource, @Nullable Transition<? super File> transition) {
                sendLoadSuccess();
                String filePath = resource.getPath();
                BitmapFactory.Options options = new BitmapFactory.Options();
                options.inJustDecodeBounds = true;
                BitmapFactory.decodeFile(filePath, options);

                //outMimeType是以--”image/png”、”image/jpeg”、”image/gif”…….这样的方式返回的
                String mimeType = options.outMimeType;

                //自定义的判断是否为GIF的工具类
                final boolean isGif = PictureMimeType.isGif(mimeType);
                final boolean isWebp = PictureMimeType.isWebp(mimeType);
                mWorkerHandler.post(new Runnable() {
                    @Override
                    public void run() {
                        if (isWebp) {
                            bitmapProvider = new WebpBitmapProvider(context, imageUrl, width, height, centerCrop);
                        } else if (isGif) {
                            bitmapProvider = new GifBitmapProvider(context, imageUrl, width, height, centerCrop);
                        } else {
                            bitmapProvider = new NormalBitmapProvider(context, imageUrl, width, height, centerCrop);
                        }
                    }
                });
            }

            @Override
            public void onLoadFailed(@Nullable Drawable errorDrawable) {
                super.onLoadFailed(errorDrawable);
                sendLoadFailed();
            }

            @Override
            public void onLoadCleared(@Nullable Drawable placeholder) {
            }
        });
    }

    private void sendLoadSuccess() {
//        if (sink != null) {
//            mainHandler.postDelayed(successRunnable, 100);
//        }
    }

    private void sendLoadFailed() {
//        if (sink != null) {
//            sink.success(2);
//        }
    }

    @Override
    public void onDispose() {
        bitmapProvider.clean();
    }
}
