
#import "NormalImageRenderWorker.h"
#import <SDWebImage/SDWebImage.h>
#import <SDWebImageWebPCoder.h>
#import "GLKit/GLKit.h"
#import <OpenGLES/ES2/glext.h>

/**
 定义顶点类型
 */
typedef struct {
    GLKVector3 positionCoord; // (X, Y, Z)
    GLKVector2 textureCoord; // (U, V)
} SenceVertex;

typedef NS_ENUM (NSUInteger, InitGLStep) {
    InitGLStepContext = 1,
    InitGLStepCreateBuffer = 2,
    InitGLStepBindBuffer = 3,
    InitGLStepDone = 4,
};

@interface NormalImageRenderWorker () {
}

@property (nonatomic, strong) EAGLContext *context;
@property (nonatomic, copy) void (^onNewFrame)(void);

@property (nonatomic, copy) NSString *url;
@property (nonatomic, assign) CGFloat aspectRatio;
@property (nonatomic, assign) CGFloat width;
@property (nonatomic, assign) CGFloat height;
@property (nonatomic, assign) BOOL centerCrop;

@property (nonatomic, strong) UIImage *image;
@property (nonatomic, strong) NSArray *imageArray;
@property (nonatomic, assign) CGFloat duration;
@property (nonatomic, strong) NSDate *startTime;
@property (nonatomic, assign) SenceVertex *vertices; // 顶点数组
@property (nonatomic, assign) BOOL needUpdate;
@property (nonatomic, assign) InitGLStep currentInitStep;
@property (nonatomic, strong) dispatch_semaphore_t lock;
@property (nonatomic, assign) BOOL animatedImageEnable;

@property (nonatomic) GLuint imageWidth;
@property (nonatomic) GLuint imageHeight;
@property (nonatomic) void *imageData;
@property (nonatomic) GLuint program;
@property (nonatomic) GLuint frameBuffer;
@property (nonatomic) GLuint depthBuffer;
@property (nonatomic) CVPixelBufferRef target;
@property (nonatomic) CVOpenGLESTextureCacheRef textureCache;
@property (nonatomic) CVOpenGLESTextureRef texture;

@end

@implementation NormalImageRenderWorker

- (instancetype)initWithCall:(FlutterMethodCall *)call onNewFrame:(void (^)(void))onNewFrame {
    self = [super init];

    if (self) {
        _onNewFrame = onNewFrame;
        _currentInitStep = InitGLStepContext;
        _lock = dispatch_semaphore_create(1);
        _animatedImageEnable = [self isAnimatedImageEnable];
        [self initCallArgument:call];
        [self initData];
    }

    return self;
}

- (void)initCallArgument:(FlutterMethodCall *)call {
    SDImageWebPCoder *webPCoder = [SDImageWebPCoder sharedCoder];
    [[SDImageCodersManager sharedManager] addCoder:webPCoder];

    _url = call.arguments[@"url"];
    _aspectRatio = [call.arguments[@"aspectRatio"] floatValue];
    _width = [call.arguments[@"width"] floatValue] * _aspectRatio;
    _height = [call.arguments[@"height"] floatValue] * _aspectRatio;
    _centerCrop = [call.arguments[@"centerCrop"] boolValue];

    NSLog(@"call Argument: \n url: %@, \n aspectRatio: %.1f, \n width:%.1f, \n height:%.1f, \n centerCrop:%ld",
            self.url, self.aspectRatio, self.width, self.height, (long) self.centerCrop);
}

- (void)initData {
    self.image = nil;
    self.imageArray = nil;
    self.duration = 0;

    UIImage *image;
    image = [[SDImageCache sharedImageCache] imageFromCacheForKey:self.url];
    if (!image) {
        [[SDWebImageDownloader sharedDownloader] downloadImageWithURL:[[NSURL alloc] initWithString:self.url] completed:^(UIImage *image, NSData *data, NSError *error, BOOL finished) {
            NSLog(@"Download image done: %@, image: %@, error: %@", self.url, image, error);
            __weak typeof(self) weakself = self;
            dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
                __strong typeof(weakself) self = weakself;
                if (!error && image) {
                    [self initImage:image];

                    // GIF图片保存后再使用会丢失images数据，无法动起来
                    if (image.sd_imageFormat != SDImageFormatGIF) {
                        [[SDImageCache sharedImageCache] storeImage:image forKey:self.url completion:nil];
                    }
                }

                if (error) {
                    self.currentInitStep = InitGLStepDone;
                }
            });
        }];

    } else {
        NSLog(@"cache image found: %@", self.url);
        [self initImage:image];
    }

    // 创建顶点数组
    self.vertices = malloc(sizeof(SenceVertex) * 4); // 4 个顶点
    self.vertices[0] = (SenceVertex) {{-1, 1, 0}, {0, 0}}; // 左上角
    self.vertices[1] = (SenceVertex) {{-1, -1, 0}, {0, 1}}; // 左下角
    self.vertices[2] = (SenceVertex) {{1, 1, 0}, {1, 0}}; // 右上角
    self.vertices[3] = (SenceVertex) {{1, -1, 0}, {1, 1}}; // 右下角

}

- (void)initImage:(UIImage *)image {
    if (self.animatedImageEnable && image.images) {
        // 有images是动图
        self.imageArray = image.images;
        self.duration = image.duration;
        self.startTime = [NSDate date];

        NSThread *thread = [[NSThread alloc] initWithTarget:self selector:@selector(run) object:nil];
        thread.name = @"NormalImageRenderWorker";
        [thread start];
        return;
    }

    self.image = image;
}

- (void)run {
    while (!self.disposed) {
        CFTimeInterval loopStart = CACurrentMediaTime();

        if (self.imageArray && self.imageArray.count > 0 && self.duration > 0.01) {
            NSTimeInterval delta = [[NSDate date] timeIntervalSinceDate:self.startTime];
            NSUInteger index = (NSUInteger) (((int) (delta * 1000) % (int) (self.duration * 1000)) * 1.0f / (int) (self.duration * 1000) * self.imageArray.count);
            UIImage *image = self.imageArray[index];
            [self drawCurrentImageToData:image];
        }

        CFTimeInterval waitDelta = 0.016 - (CACurrentMediaTime() - loopStart);
        if (waitDelta > 0) {
            [NSThread sleepForTimeInterval:waitDelta];
        }
    }
}

// 将一个顶点着色器和一个片段着色器挂载到一个着色器程序上，并返回程序的 id
- (GLuint)programWithShaderName:(NSString *)shaderName {
    // 编译两个着色器
    GLuint vertexShader = [self compileShaderWithName:shaderName type:GL_VERTEX_SHADER];
    GLuint fragmentShader = [self compileShaderWithName:shaderName type:GL_FRAGMENT_SHADER];

    // 挂载 shader 到 program 上
    GLuint program = glCreateProgram();
    glAttachShader(program, vertexShader);
    glAttachShader(program, fragmentShader);

    // 链接 program
    glLinkProgram(program);

    // 检查链接是否成功
    GLint linkSuccess;
    glGetProgramiv(program, GL_LINK_STATUS, &linkSuccess);
    if (linkSuccess == GL_FALSE) {
        GLchar messages[256];
        glGetProgramInfoLog(program, sizeof(messages), 0, &messages[0]);
        NSString *messageString = [NSString stringWithUTF8String:messages];
        NSAssert(NO, @"program链接失败：%@", messageString);
        exit(1);
    }
    return program;
}

// 编译一个 shader，并返回 shader 的 id
- (GLuint)compileShaderWithName:(NSString *)name type:(GLenum)shaderType {
    // 查找 shader 文件
    NSString *bundlePath = [[NSBundle bundleForClass:[self class]].resourcePath
            stringByAppendingPathComponent:@"/TipFlutterImageView.bundle"];
    NSBundle *bundle = [NSBundle bundleWithPath:bundlePath];
    NSString *shaderPath = [bundle pathForResource:name ofType:shaderType == GL_VERTEX_SHADER ? @"vsh" : @"fsh"]; // 根据不同的类型确定后缀名
    NSError *error;
    NSString *shaderString = [NSString stringWithContentsOfFile:shaderPath encoding:NSUTF8StringEncoding error:&error];
    if (!shaderString) {
        NSAssert(NO, @"读取shader失败");
        exit(1);
    }

    // 创建一个 shader 对象
    GLuint shader = glCreateShader(shaderType);

    // 获取 shader 的内容
    const char *shaderStringUTF8 = [shaderString UTF8String];
    int shaderStringLength = (int) [shaderString length];
    glShaderSource(shader, 1, &shaderStringUTF8, &shaderStringLength);

    // 编译shader
    glCompileShader(shader);

    // 查询 shader 是否编译成功
    GLint compileSuccess;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &compileSuccess);
    if (compileSuccess == GL_FALSE) {
        GLchar messages[256];
        glGetShaderInfoLog(shader, sizeof(messages), 0, &messages[0]);
        NSString *messageString = [NSString stringWithUTF8String:messages];
        NSAssert(NO, @"shader编译失败：%@", messageString);
        exit(1);
    }

    return shader;
}

- (void)onDraw {
    if ((self.image || self.imageArray.count > 0) && self.needUpdate) {
        if (self.currentInitStep != InitGLStepDone) {
            [self initGL];
            return;

        } else {
            [EAGLContext setCurrentContext:_context];
        }

        [self draw];
        glFlush();

        dispatch_async(dispatch_get_main_queue(), self.onNewFrame);
        self.needUpdate = NO;
    }
}

- (void)draw {
    // 设置视口尺寸
    glViewport(0, 0, (GLsizei) self.width, (GLsizei) self.height);

    // 读取纹理
    GLuint textureId = [self createTextureWithImage];

    glUseProgram(_program);

    // 获取 shader 中的参数，然后传数据进去
    GLuint positionSlot = (GLuint) glGetAttribLocation(_program, "Position");
    GLuint textureSlot = (GLuint) glGetUniformLocation(_program, "Texture");  // 注意 Uniform 类型的获取方式
    GLuint textureCoordsSlot = (GLuint) glGetAttribLocation(_program, "TextureCoords");

    // 将纹理 ID 传给着色器程序
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textureId);
    glUniform1i(textureSlot, 0);  // 将 textureSlot 赋值为 0，而 0 与 GL_TEXTURE0 对应，这里如果写 1，上面也要改成 GL_TEXTURE1

    // 创建顶点缓存
    GLuint vertexBuffer;
    glGenBuffers(1, &vertexBuffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
    GLsizeiptr bufferSizeBytes = sizeof(SenceVertex) * 4;
    glBufferData(GL_ARRAY_BUFFER, bufferSizeBytes, self.vertices, GL_STATIC_DRAW);

    // 设置顶点数据
    glEnableVertexAttribArray(positionSlot);
    glVertexAttribPointer(positionSlot, 3, GL_FLOAT, GL_FALSE, sizeof(SenceVertex), NULL + offsetof(SenceVertex, positionCoord));

    // 设置纹理数据
    glEnableVertexAttribArray(textureCoordsSlot);
    glVertexAttribPointer(textureCoordsSlot, 2, GL_FLOAT, GL_FALSE, sizeof(SenceVertex), NULL + offsetof(SenceVertex, textureCoord));

    // 开始绘制
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    // 删除顶点缓存
    glDeleteBuffers(1, &vertexBuffer);
    vertexBuffer = 0;
    glBindTexture(GL_TEXTURE_2D, 0);
}

// 通过一张图片来创建纹理
- (GLuint)createTextureWithImage {
    // 生成纹理，不能每次都glGenTextures，会导致内存泄漏，这里只能用0（default），试过1（需要通过glGenTextures）会黑屏
//    GLuint textureID;
//    glGenTextures(1, &textureID);
    glBindTexture(GL_TEXTURE_2D, 0);
    dispatch_semaphore_wait(self.lock, DISPATCH_TIME_FOREVER);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, _imageWidth, _imageHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, _imageData); // 将图片数据写入纹理缓存
    dispatch_semaphore_signal(self.lock);

    // 设置如何把纹素映射成像素
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    // 解绑
    glBindTexture(GL_TEXTURE_2D, 0);

    return 0;
}

- (void)onDispose {
    if (_vertices) {
        free(_vertices);
        _vertices = nil;
    }

    free(_imageData);

    [self deinitGL];
}

- (void)setImage:(UIImage *)image {
    if (image && self.centerCrop) {
        __weak typeof(self) weakself = self;
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
            __strong typeof(weakself) self = weakself;
            UIImage *cropImage = [self imageByCroppingImage:image bySize:CGSizeMake(self.width, self.height)];
            [self updateImageData:cropImage];
        });
    } else {
        [self updateImageData:_image];
    }
}

- (BOOL)isAnimatedImageEnable {
    // iphonex以上才支持动图，低端机很卡
    CGFloat a = 0.0f;
    if (@available(iOS 11.0, *)) {
        a = [[UIApplication sharedApplication] delegate].window.safeAreaInsets.bottom;
    }
    return a > 0;
}

- (void)updateImageData:(UIImage *)image {
    _image = image;
    if (!image) {
        return;
    }

    __weak typeof(self) weakself = self;
    dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0), ^{
        __strong typeof(weakself) self = weakself;
        [self drawCurrentImageToData:image];
    });
}

- (void)drawCurrentImageToData:(UIImage *)image {
    // 将 UIImage 转换为 CGImageRef
    CGImageRef cgImageRef = [image CGImage];
    self.imageWidth = (GLuint) CGImageGetWidth(cgImageRef);
    self.imageHeight = (GLuint) CGImageGetHeight(cgImageRef);
    CGRect rect = CGRectMake(0, 0, self.imageWidth, self.imageHeight);

    // 绘制图片
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    dispatch_semaphore_wait(self.lock, DISPATCH_TIME_FOREVER);
    if (_imageData) {
        free(_imageData);
    }
    _imageData = malloc(self.imageWidth * self.imageHeight * 4);
    CGContextRef context = CGBitmapContextCreate(_imageData, self.imageWidth, self.imageHeight, 8, self.imageWidth * 4, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big);
    CGContextTranslateCTM(context, 0, self.imageHeight);
    CGContextScaleCTM(context, 1.0f, -1.0f);
    CGColorSpaceRelease(colorSpace);
    CGContextClearRect(context, rect);
    CGContextDrawImage(context, rect, cgImageRef);
    dispatch_semaphore_signal(self.lock);

    // 释放内存
    CGContextRelease(context);

    self.needUpdate = YES;
}

- (UIImage *)imageByCroppingImage:(UIImage *)image bySize:(CGSize)size {
    double newCropWidth, newCropHeight;

    if (image.size.width * 1.0f / image.size.height < size.width * 1.0f / size.height) {
        newCropWidth = image.size.width;
        newCropHeight = (newCropWidth * size.height) / size.width;
    } else {
        newCropHeight = image.size.height;
        newCropWidth = (newCropHeight * size.width) / size.height;
    }

    double x = image.size.width / 2.0 - newCropWidth / 2.0;
    double y = image.size.height / 2.0 - newCropHeight / 2.0;

    CGRect cropRect = CGRectMake(x, y, newCropWidth, newCropHeight);
    CGImageRef imageRef = CGImageCreateWithImageInRect([image CGImage], cropRect);

    UIImage *cropped = [UIImage imageWithCGImage:imageRef];
    CGImageRelease(imageRef);

    return cropped;
}

#pragma mark - FlutterTexture

- (CVPixelBufferRef _Nullable)copyPixelBuffer {
    if (_target) {
        CVBufferRetain(_target);
    }
    return _target;
}

#pragma mark - Private

- (void)initGL {
    [self initGLContext];
    [self createBuffer];
    [self bindBuffer];
}

- (void)initGLContext {
    _context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
    [EAGLContext setCurrentContext:_context];
    _program = [self programWithShaderName:@"glsl"]; // glsl.vsh & glsl.fsh
    self.currentInitStep = InitGLStepCreateBuffer;
}

- (void)createBuffer {
    [EAGLContext setCurrentContext:_context];

    CGSize size = CGSizeMake(self.width, self.height);
    CVReturn err = CVOpenGLESTextureCacheCreate(kCFAllocatorDefault, NULL, _context, NULL, &_textureCache);
    if (err) return;

    CFDictionaryRef empty;
    CFMutableDictionaryRef attrs;
    empty = CFDictionaryCreate(kCFAllocatorDefault,
            NULL,
            NULL,
            0,
            &kCFTypeDictionaryKeyCallBacks,
            &kCFTypeDictionaryValueCallBacks);

    attrs = CFDictionaryCreateMutable(kCFAllocatorDefault, 1,
            &kCFTypeDictionaryKeyCallBacks,
            &kCFTypeDictionaryValueCallBacks);

    CFDictionarySetValue(attrs, kCVPixelBufferIOSurfacePropertiesKey, empty);
    CVPixelBufferCreate(kCFAllocatorDefault, (size_t) size.width, (size_t) size.height,
            kCVPixelFormatType_32BGRA, attrs, &_target);

    CVOpenGLESTextureCacheCreateTextureFromImage(kCFAllocatorDefault,
            _textureCache,
            _target,
            NULL, // texture attributes
            GL_TEXTURE_2D,
            GL_RGBA, // opengl format
            (GLsizei) size.width,
            (GLsizei) size.height,
            GL_BGRA, // native iOS format
            GL_UNSIGNED_BYTE,
            0,
            &_texture);

    CFRelease(empty);
    CFRelease(attrs);
    self.currentInitStep = InitGLStepBindBuffer;
}

- (void)bindBuffer {
    [EAGLContext setCurrentContext:_context];
    glBindTexture(CVOpenGLESTextureGetTarget(_texture), CVOpenGLESTextureGetName(_texture));

    glTexImage2D(GL_TEXTURE_2D,
            0, GL_RGBA,
            (GLsizei) self.width, (GLsizei) self.height,
            0, GL_RGBA,
            GL_UNSIGNED_BYTE, NULL);

    glGenRenderbuffers(1, &_depthBuffer);
    glBindRenderbuffer(GL_RENDERBUFFER, _depthBuffer);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT16, (GLsizei) self.width, (GLsizei) self.height);

    glGenFramebuffers(1, &_frameBuffer);
    glBindFramebuffer(GL_FRAMEBUFFER, _frameBuffer);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, CVOpenGLESTextureGetName(_texture), 0);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, _depthBuffer);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
        NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatus(GL_FRAMEBUFFER));
    }
    self.currentInitStep = InitGLStepDone;
}

- (void)deinitGL {
    [EAGLContext setCurrentContext:_context];

    glDeleteFramebuffers(1, &_frameBuffer);
    glDeleteFramebuffers(1, &_depthBuffer);

    if (_target) {
        CFRelease(_target);
    }
    if (_textureCache) {
        CFRelease(_textureCache);
    }
    if (_texture) {
        CFRelease(_texture);
    }
}

- (BOOL)glInitDone {
    return self.currentInitStep == InitGLStepDone;
}

@end
